#include "Perlin.h"


PerlinNoise2D::PerlinNoise2D()
{
    m_Seed  = 0;
    m_Depth = 5;

    m_Add       = 0.0f;
    m_Amplitude = 1.0f;
}

void PerlinNoise2D::SetAmplitude( FLP32 Min, FLP32 Max)
{
    m_Add       = Min;
    m_Amplitude = Max;
}

void PerlinNoise2D::SetPersistence( FLP32 Persistence)
{
    m_Persistence   = Persistence;
}

void PerlinNoise2D::SetSeed( SYSUINT Seed)
{
    m_Seed  = Seed;
}

void PerlinNoise2D::SetDepth( SYSUINT Depth)
{
    m_Depth = Depth;
}

void PerlinNoise2D::Create()
{
    FLP32   L_Range;
    FLP32   L_SummedMultiplyer;
    FLP32   L_Multiplyer;
    SYSUINT L_Count;

    L_Range             = m_Amplitude - m_Add;
    m_Add               += L_Range * 0.5f;

    // We get values between -1 and 1 out of the Rand function
    L_SummedMultiplyer  = 0.0f;
    L_Multiplyer        = 2.0f;

    for( L_Count = 0; L_Count < m_Depth; L_Count++)
    {
        L_SummedMultiplyer  += L_Multiplyer;
        L_Multiplyer        *= m_Persistence;
    }

    m_Amplitude =   L_Range / L_SummedMultiplyer;
}

FLP32 PerlinNoise2D::GetValue( FLP32 X, FLP32 Y)
{
    FLP32   L_Total         = 0.0f;
    FLP32   L_Amplitude     = m_Amplitude;
    SYSUINT L_Count;

    for( L_Count=0; L_Count<m_Depth; L_Count++)
    {
        L_Total += InterpolatedRand( X, Y) * L_Amplitude;

        L_Amplitude     *=  m_Persistence;

        X   *= 2.0f;
        Y   *= 2.0f;

        // Without this we get a strong precision loss for large values and also overflows.
        // This can't also prevent this forever but it's better than nothing.
        // A modulo operation is too time consuming but maybe a while(.. >= .. ) does a
        // better job.
        if( X >= 4096.0f )
            X -= 4096.0f;

        if( Y >= 4096.0f )
            Y -= 4096.0f;
    }

    L_Total += m_Add;

    return L_Total;
}

FLP32 PerlinNoise2D::InterpolatedRand( FLP32 X, FLP32 Y)
{
    SYSUINT L_IntX  = (SYSUINT) X;
    FLP32   L_FracX = X - L_IntX;

    SYSUINT L_IntY  = (SYSUINT) Y;
    FLP32   L_FracY = Y - L_IntY;

    FLP32   L_V1, L_V2, L_V3, L_V4, L_ResX1, L_ResX2;

    L_V1 = SmoothRand( L_IntX,     L_IntY    );
    L_V2 = SmoothRand( L_IntX + 1, L_IntY    );
    L_V3 = SmoothRand( L_IntX,     L_IntY + 1);
    L_V4 = SmoothRand( L_IntX + 1, L_IntY + 1);

    L_ResX1 = Interpolate( L_V1, L_V2, L_FracX);
    L_ResX2 = Interpolate( L_V3, L_V4, L_FracX);
    
    return Interpolate( L_ResX1, L_ResX2, L_FracY);
}

FLP32 PerlinNoise2D::Interpolate( FLP32 V1, FLP32 V2, FLP32 Step)
{
    FLP32   L_Fact;

    // This interpolation methode is faster than a cosinus function, but it's neary
    // equal to the cos-graph. I simply take a square function either in positiv
    // direction or in negativ direction starting from 1.0.

    if( Step > 0.5f )
    {
        L_Fact  = (1.0f - Step);
        L_Fact  = L_Fact * L_Fact * 2.0f;
    }
    else
    {
        L_Fact  = 1.0f - Step * Step * 2.0f;
    }

    return V1 * L_Fact + V2 * (1.0f - L_Fact);
}

FLP32 PerlinNoise2D::SmoothRand( SYSUINT X, SYSUINT Y)
{
    return Rand( X, Y );

    // The following code gives a smoother perlin noise, but it's also quite slow
    /*FLP32   L_Corners;
    FLP32   L_Sides;
    FLP32   L_Center;

    L_Corners = ( Rand(X-1, Y-1) + Rand(X+1, Y-1) + Rand(X-1, Y+1) + Rand(X+1, Y+1) ) * ( 1.0f / 16.0f );
    L_Sides   = ( Rand(X-1, Y  ) + Rand(X+1, Y  ) + Rand(X  , Y-1) + Rand(X  , Y+1) ) * ( 1.0f / 8.0f );
    L_Center  =   Rand(X  , Y  ) * ( 1.0f / 4.0f );

    return L_Center + L_Sides + L_Corners;*/
}

FLP32 PerlinNoise2D::Rand( SYSUINT X, SYSUINT Y)
{
    SYSUINT     L_Base;

    L_Base = X + Y * 57 + m_Seed;
    L_Base = (L_Base<<13) ^ L_Base;

    return ( 1.0f - ( (L_Base * (L_Base * L_Base * 15731 + 789221) + 1376312589) & 0x7fffffff) * ( 1.0f / 1073741824.0f ));
}
